
import logging

logging.basicConfig()

# Import here for static analysis to work
import silx.app.view.main
import silx.app.convert
import silx.app.test_

from silx.__main__ import main

if __name__ == "__main__":
    main()

